

//	#define TMPOUT

#include "GC_ThreadingSpace.hh"
#include "cpsp/gecode/GC_StlSetRangeIterator.hh"
#include "cpsp/gecode/GC_LatticeNeighbored2.hh"
#include "cpsp/gecode/GC_ValNearCenter.hh"
#include "cpsp/gecode/GC_RankViewSel.hh"

#include <gecode/kernel.hh>
#include <gecode/int.hh>
#include <gecode/search.hh>
#include <gecode/minimodel.hh>
#include <gecode/int/branch.hh>

#include <ell/LatticeFrame.hh>

#ifdef TMPOUT
	#include <iostream>
#endif

namespace cpsp
{
  namespace gecode
  {

  		// hilfsfunktion ... TODO : loeschen ;)
	void print(ell::IndexSet& x, const ell::LatticeFrame* latFrame) {
		std::cout<<"( ";
		for (ell::IndexSet::const_iterator it = x.begin(); it != x.end(); it++)
			std::cout <<latFrame->getPoint(*it) <<"| ";
		std::cout <<")";
		std::cout.flush();
	}

  	 		void GC_ThreadingSpace::print() const {
      std::cerr<<"\n";
      for (int i=0; i< domains.size(); i++)  {
    	const Gecode::Int::IntView vd(domains[i]);
    	Gecode::Int::ViewRanges< Gecode::Int::IntView > v(vd);
      	std::cerr <<" " <<i <<"(";
      	while(v()) {
      		if (v.min() != v.max())
      			std::cerr <<v.min() <<".." <<v.max()<<"," ;
      		else
      			std::cerr <<v.min() <<",";
      		++v;
      	}
      	std::cerr <<")\n";
      }
      std::cerr<<"\n";
		}

	GC_ThreadingSpace::GC_ThreadingSpace(const std::string *sequence,
			const ell::LatticeFrame* latFrame,
			const ell::IndexVec* neighVecs,
			SeqFeatureMap* seqFeatureMap,
			const HullLevel *hullLvl,
			HCore* hCore,
			const GC_ThreadingSymmBreaker::GlobalShiftVec* shiftVec,
			bool useSymmBreak,
			bool useNoDDS,
			NeighPropLvl neighPropLvl ) :
		domains(	this, sequence->size(),
					0, Gecode::Limits::Int::int_max), rank(sequence->size(), 999)
	{
		assert(hCore != NULL);
			// temp variables
		ell::IndexSet data;
		GC_StlSetRangeIterator dataIt;
		Gecode::IntSet dataSet;
		Gecode::IntVarArgs hVars(hCore->getSize()), pVars(domains.size() - hCore->getSize());
		int hVarI = 0, pVarI = 0;

			// domain init

		int i=0;

		for (i=0; i<domains.size(); i++) {
				// daten holen
			data = hCore->getIndexedHull(latFrame, hullLvl->at(i));
				// iterator erzeugen
			dataIt.init(&data);
				// daten einschraenken auf werte in dataIt via IntView
			Gecode::Int::IntView actDom(domains[i]);
				// macro zur sicherheit, wenn schnittmengenbildung schief laeuft
			GECODE_ME_FAIL(this, actDom.inter(this,dataIt));
				// init distinct viewvecs
			if (hullLvl->at(i) == 0) {
				hVars[hVarI++] = domains[i];
			} else {
				pVars[pVarI++] = domains[i];
			}
		}

			// p-singlet init
		std::vector<unsigned int>::const_iterator it;
			// daten holen
		data = hCore->getIndexedPSinglets(latFrame);
			// iterator erzeugen
		dataIt.init(&data);
			// domains einschraenken
		for (	it = (*seqFeatureMap)[P_SINGLET].begin();
				it != (*seqFeatureMap)[P_SINGLET].end(); it++)
		{
				// zugriffs-view auf singlet domain
			Gecode::Int::IntView actDom(domains[*it]);
				// macro zu sicherheit, wenn schnittmengenbildung schief laeuft
			GECODE_ME_FAIL(this, actDom.inter(this,dataIt));
				// rank setzen
			rank[*it] = 0;
		}

/* funzt aber bringt kaum was
	*/
			// h-singlet init
			// daten holen
		data = hCore->getIndexedHSinglets(latFrame);
			// iterator erzeugen
		dataIt.init(&data);
			// domains einschraenken
		for (	it = (*seqFeatureMap)[H_SINGLET].begin();
				it != (*seqFeatureMap)[H_SINGLET].end(); it++)
		{
				// zugriffs-view auf singlet domain
			Gecode::Int::IntView actDom(domains[*it]);
				// macro zu sicherheit, wenn schnittmengenbildung schief laeuft
			GECODE_ME_FAIL(this, actDom.inter(this,dataIt));
				// rank setzen
			rank[*it] = 0;
		}

			// constraints initialisieren

			// ALL DIFFERENT
			// Supports value (ICL_VAL, default), bounds (ICL_BND),
			// and domain-consistency (ICL_DOM)
		Gecode::distinct(this, hVars, Gecode::ICL_DOM);
		Gecode::distinct(this, pVars, Gecode::ICL_DOM);
/*
		if (!useNoDDS) {
			for (hVarI = 0; (hVarI+1) < hVars.size(); hVarI++) {
				for (i=(hVarI+1); i<hVars.size(); i++) {
//					Gecode::rel(this, hVars[hVarI], Gecode::IRT_NQ, hVars[i]);
					Gecode::rel(this, hVars[hVarI], Gecode::IRT_NQ, hVars[i], Gecode::BoolVar(this,0,1), Gecode::ICL_DOM);
//					Gecode::Int::Rel::NqSubsume<MyViewType>::post(this,hVars[hVarI], hVars[i]);
				}
			}
			for (pVarI = 0; (pVarI+1) < pVars.size(); pVarI++) {
				for (i=(pVarI+1); i<pVars.size(); i++) {
//					Gecode::rel(this, pVars[pVarI], Gecode::IRT_NQ, pVars[i]);
					Gecode::rel(this, pVars[pVarI], Gecode::IRT_NQ, pVars[i], Gecode::BoolVar(this,0,1), Gecode::ICL_DOM);
//					Gecode::Int::Rel::NqSubsume<MyViewType>::post(this,pVars[pVarI], pVars[i]);
				}
			}
		}
*/


			// init remaining rank stuff
		{
			std::string seqPattern = "HHPP";
			std::string::size_type pos = 0, shift = 1;
			while( (pos = sequence->find(seqPattern,pos)) != std::string::npos) {
				pos += shift;
				rank [pos] = 1;
			}
			seqPattern = "PPHH";
			pos = 0, shift = 2;
			while( (pos = sequence->find(seqPattern,pos)) != std::string::npos) {
				pos += shift;
				rank [pos] = 1;
			}
			if (domains.size() > 1) {
				if ((*sequence)[0] != (*sequence)[1])
					rank[0] = 1;
				if ((*sequence)[sequence->size()-1] != (*sequence)[sequence->size()-2])
					rank[sequence->size()-1] = 1;
			}
		}

			// binary lattice neighbor constraints along the sequence
		switch(neighPropLvl) {
		
		case CUSTOM_2 : {	
						for(i=1; i<domains.size(); i++) {
							GECODE_ES_FAIL(this,
							GC_LatticeNeighbored2::post(	this,
													domains[i-1], domains[i],
													neighVecs,
													Gecode::ICL_DOM,
													150
												));
						}
						break; }
		case NO_PROP : 
		default : 	//  NO NEIGHBOR PROPAGATION INSERTED
					break;
		}

			// GC_ValNearCenter lattice frame init
		GC_ValNearCenter::latFrame = latFrame;

		// branching

			// symmetry breaking
		if (useSymmBreak) {
		Gecode::ViewArray<Gecode::Int::IntView> xv(this,(Gecode::IntVarArgs)domains);
//std::cerr <<"SB : ";
			Gecode::BoolVarArray cpInit = GC_ThreadingSymmBreaker::initCpByCore(this, hCore, latFrame, shiftVec);
			GC_ThreadingSymmBreaker symmBreaker(this,latFrame, shiftVec, cpInit);

			new (this)
			Gecode::Search::SBViewValBranching
				<Gecode::Int::IntView,
				int,
				Gecode::Int::Branch::BySizeMin,
				GC_ValNearCenter,
				GC_ThreadingSymmBreaker>(this, xv,symmBreaker);
		}

			// viewarray for branching
		Gecode::ViewArray<MyViewType> xv(this,domains.size());
		for (i=0; i< domains.size(); i++)
			xv[i] = IdxIntView(domains[i], i);

        typedef Gecode::ViewValBranching<   MyViewType,
                                            int,
//                                          GC_RankViewSel<Gecode::Int::Branch::BySizeMin>,
                                            GC_RankViewSel<Gecode::Int::Branch::ByDegreeMax>,
//                                            GC_RankViewSel<Gecode::Int::Branch::BySizeMinDegreeMax>,
                                            GC_ValNearCenter
                                        > DFS;
/*                                        
        typedef Gecode::Search::DDSViewValBranching<    
        									MyViewType,
                                            int,
                                            IntDomSize,
//                                          GC_RankViewSel<Gecode::Int::Branch::BySizeMin>,
//                                            GC_RankViewSel<Gecode::Int::Branch::ByDegreeMax>,
//                                            Gecode::Int::Branch::ByDegreeMax,
                                            Gecode::Int::Branch::BySizeMinDegreeMax,
                                            GC_ValNearCenter
                                        > DDS;
*/
                                        
			// remaining search
		if (useNoDDS) {
//std::cerr <<"DFS : \n";
				// remaining
			new (this) DFS(this, xv);
/*			
		} else {
			int preSize = 0;
			for (std::vector<int>::const_iterator it=rank.begin(); it!=rank.end();it++)
				if (*it < 2)
					preSize++;
            Gecode::ViewArray<MyViewType> preViews(this,preSize);
            preSize = 0;
			for (std::vector<int>::size_type i = 0;i < rank.size(); i++)
				if (rank[i] < 2)
	                preViews[preSize++] = IdxIntView(domains[i], i);
	        new (this) DFS(this, preViews);
			
//std::cerr <<"DDS : ";
				// remaining
			new (this) DDS(this, xv);
*/
		}
	}





	GC_ThreadingSpace::GC_ThreadingSpace(bool share, GC_ThreadingSpace& toCopy):
		SuperSpace(share, toCopy), domains(), rank(toCopy.rank)
	{
		domains.update(this, share, toCopy.domains);

	}


	GC_ThreadingSpace::~GC_ThreadingSpace()
	{
	}


	SuperSpace*
	GC_ThreadingSpace::copy(bool share) {
		return new GC_ThreadingSpace(share, *this);
	}

	ell::IndexVec
	GC_ThreadingSpace::getSolution() const {
		ell::IndexVec retVec;
		for (int i=0; i<domains.size();i++) {
			if (domains[i].assigned()) {
				retVec.push_back((ell::LatticeFrame::index_type)domains[i].val());
			} else {
				std::cerr <<"Error: Variable "<<i<<" is not assigned !!! size = "<<domains[i].size()<<"\n"; 
				std::cerr <<"       Are you using DDS ?\n";
			}
		}
		return retVec;
	}


	int
	GC_ThreadingSpace::getRank(int index) const {
		assert(index >= 0 && index < (int)rank.size());
		return rank[index];
	}







  } // namespace gecode
}  // namespace cpsp


